from typing import List, Any, Optional
from datetime import datetime, timedelta
from copy import copy, deepcopy

TIME_FORMAT = "%Y-%m-%d %H:%M:%S"

class CheckPointTrackingObject:
    def __init__(self, id:str, ts:datetime, payload:Any=None):
        self.id = id
        self.ts = ts
        self.payload = deepcopy(payload)
    
    def to_json(self):
        return {
            "id": self.id,
            "ts": self.ts.strftime(TIME_FORMAT),
            "payload": deepcopy(self.payload)
        }
    
    @classmethod
    def from_json(cls, json_payload):
        return cls(
            id=json_payload["id"], 
            ts=datetime.strptime(json_payload["ts"], TIME_FORMAT), 
            payload=json_payload.get("payload")
        )
    
    def clone(self):
        return CheckPointTrackingObject(id=self.id, ts=self.ts, payload=self.payload)


class CheckPoint:
    def __init__(
        self, 
        *,
        window_start:datetime, 
        window_end: datetime, 
        max_window_length: int,
        max_step_length:int,
        min_step_length:int,
        objects: List[CheckPointTrackingObject]=[]
    ):
        assert window_start <= window_end
        assert max_window_length > 0
        assert min_step_length > 0
        assert max_step_length > 0
        assert min_step_length <= max_step_length

        self.window_start       = window_start
        self.window_end         = window_end
        self.max_window_length  = timedelta(seconds=max_window_length)
        self.max_step_length    = timedelta(seconds=max_step_length)
        self.min_step_length    = timedelta(seconds=min_step_length)
        self.objects            = copy(objects)
        self.object_ids         = set([obj.id for obj in self.objects])

    
    @classmethod
    def from_json(cls, json_payload:Optional[dict]=None):
        return cls(
            window_start        = datetime.strptime(json_payload["window_start"], TIME_FORMAT),
            window_end          = datetime.strptime(json_payload["window_end"], TIME_FORMAT),
            max_window_length   = json_payload["max_window_length"],
            max_step_length     = json_payload["max_step_length"],
            min_step_length     = json_payload["min_step_length"],
            objects             = [
                CheckPointTrackingObject.from_json(obj) for obj in json_payload.get("objects", [])
            ]
        )

    def to_json(self):
        json_payload = {
            "window_start"      : self.window_start.strftime(TIME_FORMAT),
            "window_end"        : self.window_end.strftime(TIME_FORMAT),
            "max_window_length" : int(self.max_window_length.total_seconds()),
            "max_step_length"   : int(self.max_step_length.total_seconds()),
            "min_step_length"   : int(self.min_step_length.total_seconds()),
            "objects"           : [obj.to_json() for obj in self.objects]
        }
        return json_payload
    
    def can_advance(self, new_window_end:datetime) -> bool:
        return new_window_end >= self.window_end + self.min_step_length

    def advance(
        self, 
        new_window_end:datetime, 
        objects_to_add:List[CheckPointTrackingObject]
    ) -> List[CheckPointTrackingObject]:
        if not self.can_advance(new_window_end):
            return None

        # one move cannot step longer than max_step_length
        effective_new_window_end = min(
            new_window_end, 
            self.window_end + self.max_step_length
        )

        obj_added = []
        for obj in objects_to_add:
            if obj.ts < self.window_start or obj.ts >= effective_new_window_end:
                # object not belong to the checkpoint window, ignore it
                continue
            if obj.id in self.object_ids:
                # object id already exist, ignore it
                continue
            self.objects.append(obj.clone())
            self.object_ids.add(obj.id)
            obj_added.append(obj)
        
        self.window_start = max(
            self.window_start,
            effective_new_window_end - self.max_window_length
        )
        self.window_end = effective_new_window_end
        self.objects = [obj for obj in self.objects if obj.ts >= self.window_start]
        self.object_ids     = set([obj.id for obj in self.objects])
        return obj_added



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

####################################################################################
# Command line tool to launch firebird console
# Usage:
#     fbconsole args
# The args format is the same as you pass to manage.py in a regular django app
####################################################################################

import os
import sys

def main():
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'firebirdconsole.firebirdconsole.settings')
    from django.core.management import execute_from_command_line
    execute_from_command_line(sys.argv)


if __name__ == '__main__':
    main()



from django.urls import path
from . import views
from .views import apis

pipelines = apis.PipelinesAPI()

urlpatterns = [
    path('', views.home, name='home'),
    path('pipelines/<str:pipeline_id>', views.pipeline, name='pipeline'),

    path('apis/pipelines', pipelines, name='api.pipelines'),
    path('apis/pipelines/<str:id>', pipelines, name='api.pipeline'),
]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import logging
import logging.config
import argparse
import importlib
import os
import sys
import json

LOG_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": "%(asctime)s - %(process)d - %(name)s - %(levelname)s - %(message)s"
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "standard",
            "stream"  : "ext://sys.stdout"
        },
    },
    "root": {
        "handlers": ["console"],
        "level": "DEBUG",
    },
}

# logger = logging.getLogger(__name__)

def check_args(args, names):
    for name in names:
        if getattr(args, name) is None:
            print(f"Missing option --{name.repalce('_', '-')}")
            sys.exit(2)

def main():
    parser = argparse.ArgumentParser(
        description='Pipeline Tool'
    )
    parser.add_argument(
        "action", type=str, help="Specify action",
        choices=['register', 'unregister', 'list', 'start', 'stop'],
        nargs=1
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Specify if you want to show debug logs",
    )
    parser.add_argument(
        "-cd", "--config-dir", type=str, required=False,
        help="Configuration directory"
    )
    parser.add_argument(
        "-pmn", "--pipeline-module-name", type=str, required=False,
        help="Pipeline module name"
    )
    parser.add_argument(
        "-pid", "--pipeline-id", type=str, required=False, help="pipeline ID"
    )
    parser.add_argument(
        "-r", "--replicas", type=int, default=1, required=False,
        help="replicas"
    )
    parser.add_argument(
        "-pnn", "--pipeline-namespace-name", type=str, required=False, help="pipeline namespace name"
    )
    parser.add_argument(
        "-pin", "--pipeline-image-name", type=str, required=False, help="pipeline image name"
    )
    parser.add_argument(
        "-eid", "--executor-id", type=str, required=False, help="Executor ID"
    )
    args = parser.parse_args()
    action = args.action[0]

    if args.debug:
        logging.config.dictConfig(LOG_CONFIG)
    
    config_dir = args.config_dir if args.config_dir else os.path.expanduser("~/.firebird")
    config_filename = os.path.join(config_dir, "config.json")
    if not os.path.isfile(config_filename):
        print(f"Unable to find config file from {config_filename}")
        sys.exit(1)

    with open(config_filename, "rt") as f:
        config = json.load(f)

    impl = importlib.import_module("firebird.cmd_tools.pipeline_impl")

    if action == "register":
        check_args(args, ["pipeline_namespace_name", "pipeline_image_name", "pipeline_module_name"])
        impl.register_command(
            config, 
            args.pipeline_namespace_name, 
            args.pipeline_image_name, 
            args.pipeline_module_name
        )
    if action == "unregister":
        check_args(args, ["pipeline_id"])
        impl.unregister_command(config, args.pipeline_id)
    elif action == "list":
        impl.list_command(config)
    elif action == "start":
        check_args(args, ["pipeline_id"])
        impl.start_command(config, args.pipeline_id, args.replicas)
    elif action == "stop":
        check_args(args, ["pipeline_id"])
        impl.stop_command(config, args.pipeline_id)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import logging
import logging.config
import argparse
import importlib

import json
import os
import sys

def main():
    parser = argparse.ArgumentParser(
        description='Pipeline executor'
    )
    parser.add_argument(
        "-cd", "--config-dir", type=str, default=os.path.expanduser("~/.firebird"), required=False,
        help="Configuration directory"
    )
    parser.add_argument(
        "-pid", "--pipeline-id", type=str, required=True, help="Pipeline ID"
    )
    parser.add_argument(
        "-gid", "--generator-id", type=str, help="Generator ID"
    )
    args = parser.parse_args()

    # config log
    log_config_filename = os.path.join(args.config_dir, "log_config.json")
    if os.path.isfile(log_config_filename):
        with open(log_config_filename, "rt") as f:
            LOG_CONFIG = json.load(f)
    else:
        LOG_CONFIG = {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "standard": {
                    "format": "%(asctime)s - %(process)d - %(name)s - %(levelname)s - %(message)s"
                }
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "level": "DEBUG",
                    "formatter": "standard",
                    "stream"  : "ext://sys.stdout"
                },
            },
            "root": {
                "handlers": ["console"],
                "level": "DEBUG",
            },
            "loggers": {
                "pika": {
                    "handlers": ["console"],
                    "level":    "WARNING",
                    "propagate": False
                },
                "kazoo": {
                    "handlers": ["console"],
                    "level":    "WARNING",
                    "propagate": False
                }
            }
        }
    logging.config.dictConfig(LOG_CONFIG)

    config_filename = os.path.join(args.config_dir, "config.json")
    if not os.path.isfile(config_filename):
        print(f"Unable to find config file from {config_filename}")
        sys.exit(1)

    with open(config_filename, "rt") as f:
        config = json.load(f)
    
    impl = importlib.import_module("firebird.cmd_tools.executor_impl")
    impl.execute_pipeline(
        config, 
        pipeline_id=args.pipeline_id, 
        generator_id=args.generator_id
    )


if __name__ == '__main__':
    main()



import pika
import json
from copy import copy

def get_connection( *, username, password, host, port, heartbeat):
    credentials = pika.PlainCredentials(username, password)
    return pika.BlockingConnection(pika.ConnectionParameters(
        host=host,
        port=port,
        credentials=credentials,
        heartbeat=heartbeat
    ))


class RabbitMQ:
    def __init__(self, *, connection, topic):
        self.topic = topic
        self.error_topic = f"{topic}-error"

        self.stop_requested = False  # has the client asked to stop consuming data?
        self.connection = connection
        self.channel = self.connection.channel()
        self.channel.basic_qos(prefetch_count=1)

    def stop_consume(self):
        self.stop_consume = True
    
    def close(self):
        if self.channel is not None:
            self.channel.close()
            self.channel = None
        if self.connection is not None:
            self.connection.close()
            self.connection = None

    def initialize(self):
        self.channel.queue_declare(queue=self.topic, durable=True)
        self.channel.queue_declare(queue=self.error_topic, durable=True)

    def produce(self, json_body):
        body = json.dumps(json_body).encode('utf-8')
        self.channel.basic_publish(
            exchange='', 
            routing_key=self.topic, 
            body=body,
            properties=pika.BasicProperties(
                delivery_mode = 2, # make message persistent
            )
        )

    def produce_error(self, json_body, exception):
        effective_json_body = copy(json_body)
        if "failed_reason" not in effective_json_body:
            effective_json_body["failed_reason"] = []
        
        effective_json_body["failed_reason"].append(str(exception))
        body = json.dumps(effective_json_body).encode('utf-8')

        self.channel.basic_publish(
            exchange='', 
            routing_key=self.error_topic, 
            body=body,
            properties=pika.BasicProperties(
                delivery_mode = 2, # make message persistent
            )
        )
    
    def consume(self, on_message, quit_requested):
        # on_message: called when there is a message coming
        for method_frame, properties, body in self.channel.consume(
            self.topic, auto_ack=False, exclusive=False,
            inactivity_timeout = 10
        ):
            if quit_requested.value:
                # we are not acking the message so it will not be lost
                break
            
            if method_frame is None:
                # no message
                continue

            json_body = json.loads(body.decode('utf-8'))

            try:
                on_message(json_body)
            except Exception as e:
                # on failure, put in the error queue
                self.produce_error(json_body, e)
            finally:
                self.channel.basic_ack(method_frame.delivery_tag)




import logging

from .api_base import RESTAPIBase, ServerInternalError

import os
from django.http  import HttpRequest
from django.conf import settings
from firebird import zkdb, ZKError
import tempfile
import xml.dom.minidom
import graphviz
from django.core.exceptions import BadRequest
from firebird.libs.k8 import K8ACCESSOR

logger = logging.getLogger(__name__)


def get_generator_ids(pipeline):
    generator_ids = []
    for node in pipeline["info"]["nodes"]:
        input_port_count = 0
        for port in node["ports"]:
            if port["type"] == "INPUT": # hack, better use PortType enum
                input_port_count += 1
                break
        if input_port_count == 0:
            generator_ids.append(node["id"])
    return generator_ids

class PipelinesAPI(RESTAPIBase):
    def list(self, request:HttpRequest, **kwargs):
        with zkdb(**settings.FIREBIRD_CONFIG['zookeeper']) as db:
            pipelines = db.get_pipelines()
        return {"pipelines": pipelines}

    def get(self, request:HttpRequest, id:str, **kwargs):
        with zkdb(**settings.FIREBIRD_CONFIG['zookeeper']) as db:
            error, pipeline = db.get_pipeline(id)
            if error != ZKError.OK:
                raise BadRequest(f"Cannot get pipeline {id}")

        filename = None
        svg_filename = None
        with tempfile.NamedTemporaryFile(prefix="firebird-svg-", delete=False) as f:
            filename = f.name
            svg_filename = f"{filename}.svg"

        svgs = []
        for rankdir in ["LR", "TB"]:
            try:
                edges = set()
                g = graphviz.Digraph(format="svg")
                g.attr(bgcolor="transparent")
                g.attr(rankdir=rankdir)
                for node in pipeline["info"]["nodes"]:
                    g.node(node['id'], node['title'], shape="box", href="#", style="filled", fillcolor="green")

                    for port in node["ports"]:
                        if port["type"] == "OUTPUT":
                            for connected_port in port["connected_ports"]:
                                next_node_id, _ = connected_port.split(":")
                                edges.add((node["id"], next_node_id))
                for src_node_id, next_node_id in edges:
                    g.edge(src_node_id, next_node_id)
                g.render(filename=filename)

                doc = xml.dom.minidom.parse(svg_filename)
                svg = doc.documentElement.toxml()
                svgs.append(svg)
            finally:
                for fn in [filename, svg_filename]:
                    if os.path.isfile(fn):
                        os.remove(fn)

        return {
            "pipeline": pipeline,
            "svg_lr": svgs[0],
            "svg_tb": svgs[1],
        }

    def start(self, request:HttpRequest, id:str, **kwargs):
        with zkdb(**settings.FIREBIRD_CONFIG['zookeeper']) as db:
            error, pipeline = db.get_pipeline(id)

            if error != ZKError.OK:
                raise BadRequest(f"Unable to get pipeline, error is {error}")

            if pipeline['is_running']:
                raise BadRequest(f"Pipeline is already running!")
        
            K8ACCESSOR.apply(
                template_name="pipeline.yaml",
                context={
                    "pipeline_namespace_name": pipeline["namespace_name"],
                    "pipeline_image_name": pipeline["image_name"],
                    "pipeline_id": id,
                    "replicas": 1,
                    "generator_ids": get_generator_ids(pipeline)
                }
            )
            error = db.set_pipeline_is_running(id, True)
            if error != ZKError.OK:
                raise ServerInternalError(f"Unable to change pipeline to running status, error is {error}")



    def stop(self, request:HttpRequest, id:str, **kwargs):
        with zkdb(**settings.FIREBIRD_CONFIG['zookeeper']) as db:
            error, pipeline = db.get_pipeline(id)

            if error != ZKError.OK:
                raise BadRequest(f"Unable to get pipeline, error is {error}")

            if not pipeline['is_running']:
                raise BadRequest(f"Pipeline is not running!")


            for generator_id in get_generator_ids(pipeline):
                name = f"firebird-pipeline--{id}--{generator_id}"
                K8ACCESSOR.delete_statefulset(namespace=pipeline["namespace_name"], name=name)

            name = f"firebird-pipeline--{id}"
            K8ACCESSOR.delete_deployment(namespace=pipeline["namespace_name"], name=name)

            error = db.set_pipeline_is_running(id, False)
            if error != ZKError.OK:
                raise BadRequest(f"Unable to change pipeline to stopped status, error is {error}")



from django.shortcuts import render
import json

def render_application(request, *, 
    scripts, sub_title, app_context=None, init_menu_key=None,
    status=200
):
    context={
        'user': request.user,
        'scripts': scripts,
        'sub_title': sub_title
    }

    if app_context is not None:
        context['app_context'] = json.dumps(app_context)
    
    if init_menu_key is not None:
        context['init_menu_key'] = init_menu_key

    return render(
        request,
        'application.html',
        context=context,
        status=status
    )




def PatchRequestMiddleware(get_response):

    def middleware(request):
        if request.method == "POST" and request.headers.get('X-Use-Method') == "PATCH":
            request.method = "PATCH"
        response = get_response(request)
        return response

    return middleware


#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from typing import Dict, Optional
import logging

###################################################################
# the main process need to monitor child process
# - if main process got SIGTERM (killed), it will set quit_requested to True
#   this variable has been shared with child processes, so child process can check it and quit
#   accordingly
# - if a child process is terminated. The main process knows it
#   TODO: the main process can check the child process exit code and retry it if needed
# - if a child process got SIGTERM, it will quit.
###################################################################
logger = logging.getLogger(__name__)

import sys
import importlib
import signal
from multiprocessing import Process, Value
from uuid import uuid4
from firebird import zkdb, Node, ZKError

from firebird.rabbitmq import get_connection, RabbitMQ

__EXECUTOR_CONTEXT = {
    'quit_requested': Value('b', False)
}
    
def __request_executor_shutdown(signal_umber, frame):
    _, _ = signal_umber, frame
    __EXECUTOR_CONTEXT['quit_requested'].value = True
    logger.info("executor: Got SIGTERM")


def executor_main(config:dict, pipeline_module_name:str, pipeline_id:str, node: Optional[Node]):
    signal.signal(signal.SIGTERM, __request_executor_shutdown)  # regular kill command
    quit_requested = __EXECUTOR_CONTEXT['quit_requested']

    mq = RabbitMQ(
        connection = get_connection(**config["rabbitmq"]),
        topic = pipeline_id
    )
    pipeline = importlib.import_module(pipeline_module_name).get_pipeline(mq)
    
    if node is None:
        # for puller
        pipeline.message_loop(quit_requested)
    else:
        # for generator
        generator = pipeline[node.id]
        generator.pump(quit_requested)

def execute_pipeline(
    config:dict, 
    *, 
    pipeline_id:str, 
    generator_id:Optional[str]
):
    logger.info(f"execute_pipeline: enter")
    logger.info(f"execute_pipeline: pipeline_id={pipeline_id}, generator_id={generator_id}")
    zk_config = config['zookeeper']

    with zkdb(**zk_config) as db:
        error, pipeline = db.get_pipeline(pipeline_id)
        if error != ZKError.OK:
            logger.info(f"execute_pipeline: uable to get pipeline {pipeline_id}, error is {error}!")
            logger.info(f"execute_pipeline: exit")
            sys.exit(1)

        pipeline_module_name = pipeline['module']
        logger.info(f"execute_pipeline: pipeline_module_name={pipeline_module_name}")

        # try to import the pipeline
        pipeline = importlib.import_module(pipeline_module_name).get_pipeline(None)
        logger.info(f"execute_pipeline: pipeline acquired")

        executor_id = str(uuid4())
        logger.info(f"execute_pipeline: executor_id={executor_id}")
        error = db.register_executor(pipeline_id, executor_id, generator_id)
        if error != ZKError.OK:
            logger.info(f"execute_pipeline: uable to register executor {executor_id}, error is {error}!")
            logger.info(f"execute_pipeline: exit")
            sys.exit(1)

        logger.info(f"execute_pipeline: executor {executor_id} registered")

    try:
        node = None if generator_id is None else pipeline[generator_id]
        executor_main(config, pipeline_module_name, pipeline_id, node)
    finally:
        with zkdb(**zk_config) as db:
            error = db.unregister_executor(pipeline_id, executor_id)
            if error == ZKError.OK:
                logger.info(f"execute_pipeline: executor {executor_id} unregistered")
            else:
                logger.info(f"execute_pipeline: uable to unregister executor {executor_id}, error is {error}!")
            logger.info(f"execute_pipeline: exit")




from django.http  import HttpRequest, HttpResponse
from firebirdconsole.ui.view_tools import render_application

def pipeline(request:HttpRequest, pipeline_id:str)->HttpResponse:
    return render_application(
        request, 
        scripts = ['/static/js-bundle/pipeline.js'], 
        sub_title='home',
        app_context = {
            "pipeline_id": pipeline_id
        }
    )



from django import template
import json

register = template.Library()


# convert object to json
@register.filter(name='user_to_json')
def user_to_json(user):
    if not user.is_authenticated:
        payload = {
            'is_authenticated': False
        }
    else:
        payload = {
            'is_authenticated': True,
            'id': user.id,
            'username': user.username,
            'first_name': user.first_name,
            'last_name': user.last_name,
            'email': user.email,
            'is_staff': user.is_staff,
            'is_superuser': user.is_superuser
        }
    return json.dumps(payload)



from typing import Any
from pathlib import Path
import jinja2

JINJA_ENV   = jinja2.Environment()
BASE_DIR    = Path(__file__).resolve().parent.parent

def render_template(template_name:str, context:Any)->str:
    template_filename = BASE_DIR / f"resources/templates/{template_name}"

    with open(template_filename, "rt") as f:
        content = f.read()
        template = JINJA_ENV.from_string(content)
        return template.render(context)




from typing import Dict, Callable, Any
import tempfile
import os
from kubernetes import client,config, utils
from firebird.libs import render_template

class K8Accessor:
    def __init__(self):
        config.load_kube_config()
        self.api = client.AppsV1Api()
        self.client = client.ApiClient()
    
    def delete_deployment(self, *, namespace:str, name:str) -> None:
        resp = self.api.delete_namespaced_deployment(
            name=name,
            namespace=namespace,
            pretty=True,
            body=client.V1DeleteOptions(
                propagation_policy="Foreground", grace_period_seconds=300
            ),
        )
        print(resp)
    
    def delete_statefulset(self, *, namespace:str, name:str) -> None:
        resp = self.api.delete_namespaced_stateful_set(
            name=name,
            namespace=namespace,
            pretty=True,
            body=client.V1DeleteOptions(
                propagation_policy="Foreground", grace_period_seconds=300
            ),
        )
        print(resp)


    def apply(self, *, template_name:str, context: Any) -> None:
        to_apply = render_template(template_name, context)
        with tempfile.NamedTemporaryFile(mode='wt', delete=False) as tf:
            tf.write(to_apply)
        try:
            print("Use following deployment:")
            print(os.linesep)
            print(os.linesep)
            print(to_apply)
            print(os.linesep)
            print(os.linesep)
            utils.create_from_yaml(self.client, tf.name, verbose=True)
        finally:
            os.remove(tf.name)


K8ACCESSOR = K8Accessor()



from django.http  import HttpRequest, HttpResponse
from firebirdconsole.ui.view_tools import render_application

def home(request:HttpRequest)->HttpResponse:
    return render_application(
        request, 
        scripts = ['/static/js-bundle/home.js'], 
        sub_title='home'
    )



###################################################
# A Simple Rest API Framework
# Rest API Implementation must derive from RESTAPIBase
# - Rest API implementation can implement list, get, update, create, delete method
# - Rest API implementation can implement custom method as well
# - Rest API implementation http methods MUST return JSON object or throw recognized exception
# - Any unrecognized exception will be considered as server internal error and result
#   in HTTP 500 response
###################################################

import logging

from django.http import JsonResponse
from django.http  import HttpRequest, HttpResponse
from django.core.exceptions import BadRequest

logger = logging.getLogger(__name__)

class ServiceUnavailableException(Exception):
    pass

class ServerInternalError(Exception):
    pass

class RESTAPIBase:
    def list(self, request:HttpRequest, **kwargs):
        # returns JSON object
        raise BadRequest('LIST is not supported')

    def get(self, request:HttpRequest, id:str, **kwargs):
        # returns JSON object
        raise BadRequest('GET is not supported')

    def update(self, request:HttpRequest, id:str, partial:bool=False, **kwargs):
        # returns JSON object
        raise BadRequest('UPDATE is not supported')
    
    def create(self, request:HttpRequest, **kwargs):
        # returns JSON object
        raise BadRequest('CREATE is not supported')

    def delete(self, request:HttpRequest, id:str, **kwargs):
        # returns JSON object
        raise BadRequest('DELETE is not supported')

    def call_unsafe(self, request:HttpRequest, **kwargs):
        resp_json = None
        if request.method == "GET":
            id = kwargs.pop("id", None)
            if id is None:
                resp_json = self.list(request, **kwargs)
            else:
                resp_json = self.get(request, id, **kwargs)
        elif request.method == "PUT":
            id = kwargs.pop("id", None)
            assert id is not None
            resp_json = self.update(request, id, partial=False, **kwargs)
        elif request.method == "PATCH":
            id = kwargs.pop("id", None)
            assert id is not None
            resp_json = self.update(request, id, partial=True, **kwargs)
        elif request.method == "POST":
            method = request.headers.get("X-Use-Method")
            if method is None:
                resp_json = self.create(request, **kwargs)
            else:
                try:
                    m = getattr(self, method)
                    resp_json = m(request, **kwargs)
                except AttributeError:
                    raise BadRequest(f'{method} is not supported')
        elif request.method == "DELETE":
            id = kwargs.pop("id", None)
            assert id is not None
            resp_json = self.delete(request, id, **kwargs)
        else:
            raise BadRequest(f'{request.method} is not recognized')
        return JsonResponse(resp_json, safe=False)

    def __call__(self, request:HttpRequest, **kwargs):
        try:
            return self.call_unsafe(request, **kwargs)
        except BadRequest as e:
            return HttpResponse(str(e), content_type="text/plain", status=400, reason=str(e))
        except ServiceUnavailableException as e:
            return HttpResponse(str(e), content_type="text/plain", status=503, reason=str(e))
        except Exception as e:
            return HttpResponse(str(e), content_type="text/plain", status=500, reason=str(e))





from typing import Optional, Any, Tuple
from enum import Enum
import json
import os
from datetime import datetime
from contextlib import contextmanager
from kazoo.client import KazooClient

@contextmanager
def zkdb(**kwargs):
    zk = KazooClient(**kwargs)
    db = ZKDatabase(zk)
    zk.start()
    try:
        zk.ensure_path("/pipelines")
        zk.ensure_path("/firebird-global-lock")
        # TODO: Monolithic lock is not ideal, only for MVP
        with zk.Lock("/firebird-global-lock", "zkdb"):
            yield db
    finally:
        zk.stop()
        zk.close()

class ZKError(Enum):
    OK = 1                          # Operation completed successfully
    PIPELINE_ALREADY_REGISTERED = 2
    PIPELINE_NOT_REGISTERED     = 3
    PIPELINE_IS_RUNNING         = 4
    PIPELINE_HAS_EXECUTORS      = 5
    EXECUTOR_ALREADY_EXIST      = 6
    EXECUTOR_NOT_FOUND          = 7

class ZKDatabase:
    """
    Manage zookeeper database
    """
    def __init__(self, zk):
        self.zk = zk
    
    def register_pipeline(
        self, 
        pipeline_id:str, 
        pipeline_namespace_name:str, 
        pipeline_image_name:str, 
        pipeline_module_name:str, 
        pipeline_info:str
    ) -> ZKError:
        """
        Register a pipeline
        """
        pipeline_path = f"/pipelines/{pipeline_id}"
        if self.zk.exists(pipeline_path):
            return ZKError.PIPELINE_ALREADY_REGISTERED
        
        self.zk.ensure_path(pipeline_path)
        self.zk.create(f"{pipeline_path}/info", json.dumps(pipeline_info).encode("utf-8"))
        self.zk.create(f"{pipeline_path}/module", pipeline_module_name.encode("utf-8"))
        self.zk.create(f"{pipeline_path}/namespace_name", pipeline_namespace_name.encode("utf-8"))
        self.zk.create(f"{pipeline_path}/image_name", pipeline_image_name.encode("utf-8"))
        self.zk.create(f"{pipeline_path}/is_running", b'0')
        return ZKError.OK

    def unregister_pipeline(self, pipeline_id:str) -> ZKError:
        """
        Unregister a pipeline
        """
        pipeline_path = f"/pipelines/{pipeline_id}"
        if not self.zk.exists(pipeline_path):
            return ZKError.PIPELINE_NOT_REGISTERED
        
        # You can only unregister a pipeline if it is not running
        value, version = self.zk.get(f"{pipeline_path}/is_running")
        if value != b'0':
            return ZKError.PIPELINE_IS_RUNNING

        executors_path = f"{pipeline_path}/executors"
        if self.zk.exists(executors_path) and len(self.zk.get_children(executors_path)) > 0:
            return ZKError.PIPELINE_HAS_EXECUTORS
        
        self.zk.delete(pipeline_path, recursive=True)
        return ZKError.OK

    def get_pipelines(self):
        """
        Get information for all pipelines
        """
        pipelines = []
        for pipeline_id in self.zk.get_children("/pipelines"):
            status, pipeline = self.get_pipeline(pipeline_id)
            assert status == ZKError.OK
            assert pipeline is not None
            pipelines.append(pipeline)
        return pipelines

    def get_executor(self, pipeline_id:str, executor_id:str) -> Tuple[ZKError, Any]:
        """
        Get information for an executor
        """
        executor_path = f"/pipelines/{pipeline_id}/executors/{executor_id}"
        if not self.zk.exists(executor_path):
            return ZKError.EXECUTOR_NOT_FOUND, None
        v, _ = self.zk.get(f"{executor_path}/info")
        executor_info = json.loads(v.decode("utf-8"))
        executor_info['id'] = executor_id
        return ZKError.OK, {"info": executor_info}

    def get_pipeline(self, pipeline_id:str) -> Tuple[ZKError, Any]:
        """
        Get information for a pipeline
        """
        pipeline_path = f"/pipelines/{pipeline_id}"
        if not self.zk.exists(pipeline_path):
            return ZKError.PIPELINE_NOT_REGISTERED, None

        value, _ = self.zk.get(f"{pipeline_path}/info")
        pipeline_info = json.loads(value.decode("utf-8"))
        value, _ = self.zk.get(f"{pipeline_path}/module")
        module = value.decode("utf-8")
        value, _ = self.zk.get(f"{pipeline_path}/namespace_name")
        namespace_name = value.decode("utf-8")
        value, _ = self.zk.get(f"{pipeline_path}/image_name")
        image_name = value.decode("utf-8")
        value, _ = self.zk.get(f"{pipeline_path}/is_running")
        is_running = value == b'1'

        executor_ids_path = f"/pipelines/{pipeline_id}/executors"
        executors = []
        if self.zk.exists(executor_ids_path):
            for executor_id in self.zk.get_children(executor_ids_path):
                error, executor = self.get_executor(pipeline_id, executor_id)
                assert error == ZKError.OK
                assert executor is not None
                executors.append(executor)
        
        return ZKError.OK, {
            "info": pipeline_info,
            "module": module,
            "image_name": image_name,
            "namespace_name": namespace_name,
            "is_running": is_running,
            "executors": executors
        }


    def register_executor(
        self, 
        pipeline_id:str, 
        executor_id:str, 
        generator_id: Optional[str]
    ) ->ZKError:
        """
        Register an executor
        """
        if not self.zk.exists(f"/pipelines/{pipeline_id}"):
            return ZKError.PIPELINE_NOT_REGISTERED
    
        executors_path = f"/pipelines/{pipeline_id}/executors"
        self.zk.ensure_path(executors_path)

        executor_path = f"{executors_path}/{executor_id}"
        if self.zk.exists(executor_path):
            return ZKError.EXECUTOR_ALREADY_EXIST
        
        self.zk.ensure_path(executor_path)

        executor_info = {
            "start_time": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
            "pid": os.getpid(),
            "generator_id": generator_id
        }
        self.zk.create(f"{executor_path}/info", json.dumps(executor_info).encode("utf-8"))
        return ZKError.OK

     
    def unregister_executor(self, pipeline_id:str, executor_id:str) -> ZKError:
        """
        Unregister an executor
        """
        executor_path = f"/pipelines/{pipeline_id}/executors/{executor_id}"
        if not self.zk.exists(executor_path):
            return ZKError.EXECUTOR_NOT_FOUND
        self.zk.delete(executor_path, recursive=True)
        return ZKError.OK

    def set_pipeline_is_running(self, pipeline_id:str, is_running:bool) -> ZKError:
        pipeline_path = f"/pipelines/{pipeline_id}"
        if not self.zk.exists(pipeline_path):
            return ZKError.PIPELINE_NOT_REGISTERED
        value, _ = self.zk.get(f"{pipeline_path}/is_running")
        target_value = b'1' if is_running else b'0'
        if value != target_value:
            self.zk.set(f"{pipeline_path}/is_running", target_value)
        return ZKError.OK



from abc import ABC, abstractmethod
from typing import List, Any, Optional, Dict, Set, Union, Tuple
from enum import Enum
from multiprocessing.sharedctypes import Synchronized

from firebird.rabbitmq import RabbitMQ
########################################################################
# Node:         a node of a pipeline
# Sink:         a node without output ports
# Generator:    a node without input ports
# Port:         input port can connect to output port
#
# User do not derive class from Port
########################################################################

DEFAULT_INPUT_PORT_ID   = "input"
DEFAULT_OUTPUT_PORT_ID  = "output"

class PortType(Enum):
    INPUT = 1
    OUTPUT = 2
    
class Node(ABC):
    # Each input port has a unique name among input ports
    # Each output port has a unique name among output ports
    # Sink does not have output ports
    # Source does not have input ports
    def __init__(self, 
        *, 
        id:str,
        pipeline:"Pipeline", 
        input_port_ids:List[str]=[DEFAULT_INPUT_PORT_ID], 
        output_port_ids:List[str]=[DEFAULT_OUTPUT_PORT_ID],
        title:str="",
        description:str="",
    ):
        # validate input port ids and output port ids
        all_port_id_set = set(input_port_ids + output_port_ids)
        if len(all_port_id_set) < len(input_port_ids) + len(output_port_ids):
            raise Exception(f"Duplicate port ids found!")

        self._ports_dict:Dict[str, "Port"] = {}
        self._id:str = id
        self._pipeline:"Pipeline" = pipeline
        self._description:str = description
        if title:
            self._title = title
        else:
            self._title = id
        
        self._input_port_ids:Tuple["Node"] = tuple(input_port_ids)
        self._output_port_ids:Tuple["Node"] = tuple(output_port_ids)

        for port_id in input_port_ids:
            self._ports_dict[port_id] = Port(PortType.INPUT, port_id, self)

        for port_id in output_port_ids:
            self._ports_dict[port_id] = Port(PortType.OUTPUT, port_id, self)
        
        pipeline._add_node(self)
    
    def to_json(self):
        ret = {
            "id": self.id,
            "title": self.title,
            "description": self.description,
            "ports": [port.to_json() for port in self._ports_dict.values()]
        }
        return ret
    
    def __getitem__(self, port_id:str) -> Optional["Port"]:
        return self._get_port(port_id)

    @property
    def id(self) -> str:
        return self._id

    @property
    def pipeline(self) -> "Pipeline":
        return self._pipeline

    @property
    def description(self) -> str:
        return self._description

    @property
    def title(self) -> str:
        return self._title

    @property
    def input_port_ids(self) -> Tuple["Node"]:
        return self._input_port_ids

    @property
    def output_port_ids(self) -> Tuple["Node"]:
        return self._output_port_ids
    
    @property
    def input(self) -> Optional["Port"]:
        # return default input port
        return self._get_port(DEFAULT_INPUT_PORT_ID)
    
    @property
    def output(self) -> Optional["Port"]:
        # return default output port
        return self._get_port(DEFAULT_OUTPUT_PORT_ID)
   
    def _get_port(self, id:str) -> Optional["Port"]:
        # get input port by name
        return self._ports_dict.get(id)

    @abstractmethod
    def on_message(self, port_id:str, data:Any):
        # data arrives form input port with port_id
        pass

    def __lshift__(self, other:Union["Node", "Port", Tuple[Union["Node", "Port"]], List[Union["Node", "Port"]]]):
        # for self << other
        assert other is not None
        if isinstance(other, tuple) or isinstance(other, list):
            for t in other:
                t._connect(self)
            return other

        other._connect(self)
        return other

    def __rshift__(self, other:Union["Node", "Port", Tuple[Union["Node", "Port"]], List[Union["Node", "Port"]]]):
        # for self >> other
        assert other is not None
        if isinstance(other, tuple) or isinstance(other, list):
            for t in other:
                self._connect(t)
            return other

        self._connect(other)
        return other

    def __rlshift__(self, others:Union[Tuple[Union["Node", "Port"]], List[Union["Node", "Port"]]]):
        # for other << self
        for other in others:
            self._connect(other)
        return self

    def __rrshift__(self, others:Union[Tuple[Union["Node", "Port"]], List[Union["Node", "Port"]]]):
        # for other >> self
        for other in others:
            other._connect(self)
        return self

    def _connect(self, other: Union["Node", "Port"]):
        src = self.output
        if isinstance(other, Port):
            dst = other
        else:
            assert isinstance(other, Node)
            dst = other.input
        
        src._connect(dst)
    
    def emit(self, json_data:Any, port_id:str=DEFAULT_OUTPUT_PORT_ID):
        self._get_port(port_id).emit(json_data)
    
    def show_info(self, prefix=""):
        print(f"{prefix}node: id={self.id}")
        for port_id in self.input_port_ids:
            self._get_port(port_id).show_info(prefix=prefix+"    ")
        for port_id in self.output_port_ids:
            self._get_port(port_id).show_info(prefix=prefix+"    ")
    
    def is_generator(self):
        return isinstance(self, Generator)

class Sink(Node):
    def __init__(self, 
        *,
        id:str,
        pipeline:"Pipeline", 
        input_port_ids:List[str]=[DEFAULT_INPUT_PORT_ID],
        title:str="",
        description:str="",
    ):
        super().__init__(
            id=id, 
            pipeline=pipeline, 
            input_port_ids=input_port_ids, 
            output_port_ids=[], 
            title=title,
            description=description
        )

class Generator(Node):
    def __init__(self, 
        *,
        id:str,
        pipeline:"Pipeline", 
        output_port_ids:List[str]=[DEFAULT_OUTPUT_PORT_ID],
        title:str="",
        description:str="",
    ):
        super().__init__(
            id=id, 
            pipeline=pipeline, 
            input_port_ids=[], 
            output_port_ids=output_port_ids,
            title=title,
            description=description
        )

    def on_message(self, name:str, data:Any):
        raise Exception("Generator cannot process data")

    @abstractmethod
    def pump(self, quit_requested: Synchronized):
        """
        Let the source to collect data and emit data

        quit_requested: a boolean shared value
        """
        pass


class Port:
    def __init__(self, type:PortType, id:str, owner:Node):
        self._id:str        = id
        self._type:PortType = type
        self._owner:Node    = owner
        self._connected_ports:Set["Port"] = set()

    def to_json(self):
        ret = {
            "id": self.id,
            "type": self.type.name,
            "connected_ports": [f"{port.owner.id}:{port.id}" for port in self._connected_ports],
        }
        return ret

    @property
    def id(self) -> str:
        return self._id

    @property
    def type(self) -> PortType:
        return self._type

    @property
    def owner(self) -> Node:
        return self._owner

    def __lshift__(self, other:Union["Port", Tuple["Port"], List["Port"]]):
        # for self << other
        assert other is not None
        if isinstance(other, tuple) or isinstance(other, list):
            for t in other:
                t._connect(self)
            return other

        other._connect(self)
        return other

    def __rshift__(self, other:Union["Port", List["Port"]]):
        # for self >> other
        assert other is not None
        if isinstance(other, tuple) or isinstance(other, list):
            for t in other:
                self._connect(t)
            return other

        self._connect(other)
        return other

    def __rlshift__(self, others:Union[Tuple["Port"], List["Port"]]):
        # for other << self
        for other in others:
            self._connect(other)
        return self


    def __rrshift__(self, others:Union[Tuple["Port"], List["Port"]]):
        # for other >> self
        for other in others:
            other._connect(self)
        return self

    def _connect(self, other: Union["Node", "Port"]):
        if isinstance(other, Port):
            dst = other
        else:
            assert isinstance(other, Node)
            dst = other.input
        
        if self.type != PortType.OUTPUT:
            raise Exception("Source port MUST be output port!")
        if dst.type != PortType.INPUT:
            raise Exception("Destination port MUST be input port!")

        dst._connected_ports.add(self)
        self._connected_ports.add(dst)

    
    def emit(self, json_data:Any):
        if self.type != PortType.OUTPUT:
            raise Exception("Can only emit data to output port!")

        pipeline = self.owner.pipeline

        for port in self._connected_ports:
            pipeline.mq.produce({
                "from": {
                    "node": self.owner.id,
                    "port": self.id
                },
                "to": {
                    "node": port.owner.id,
                    "port": self.id
                },
                "payload": json_data
            })
    
    def show_info(self, prefix=""):
        t = ",".join([f"{port.owner.id}:{port.id}" for port in self._connected_ports])
        if self.type == PortType.INPUT:
            connect_str = f"from: {t}"
        else:
            connect_str = f"to: {t}"
        print(f"{prefix}port: id={self.id}, type={self.type}, {connect_str}")


    
class Pipeline:
    def __init__(self, *, id:str, mq:Optional[RabbitMQ], title:str="", description:str=""):
        self._id:str = id
        self._description:str = description
        if title:
            self._title:str = title
        else:
            self._title:str = id
        self._mq:RabbitMQ  = mq
        self._node_dict:Dict[str, Node] = {}

    def to_json(self):
        ret = {
            "id": self.id,
            "title": self.title,
            "description": self.description,
            "nodes": [node.to_json() for node in self._node_dict.values()]
        }
        return ret

    @property
    def id(self) -> str:
        return self._id

    @property
    def description(self) -> str:
        return self._description

    @property
    def title(self) -> str:
        return self._title

    @property
    def mq(self) -> Optional[RabbitMQ]:
        return self._mq

    def message_loop(self, quit_requested: Synchronized):
        self.mq.consume(self.on_message, quit_requested)
    
    def on_message(self, envelope):
        node = self._node_dict[envelope['to']['node']]
        node.on_message(
            envelope["to"]["port"],
            envelope["payload"]
        )

    def _add_node(self, node:Node):
        # add a node to the pipeline
        if node.id in self._node_dict:
            raise Exception(f"Node {node.id} already exist!")
        self._node_dict[node.id] = node
    
    
    def __getitem__(self, node_id:str) -> Optional[Node]:
        return self._node_dict.get(node_id)

    @property
    def nodes(self):
        return tuple(self._node_dict.values())

    def show_info(self, prefix=""):
        print(f"{prefix}pipeline: id={self.id}")
        for node_id, node in self._node_dict.items():
            assert node.id == node_id
            node.show_info(prefix=prefix+"    ")




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from typing import Dict, Callable
import logging
import logging.config
import importlib
from uuid import uuid4
from firebird.rabbitmq import get_connection, RabbitMQ
from firebird import zkdb, ZKError
from firebird.libs import render_template
from firebird.libs.k8 import K8ACCESSOR

logger = logging.getLogger(__name__)

def register_command(config:dict, pipeline_namespace_name:str, pipeline_image_name:str, pipeline_module_name:str):
    pipeline = importlib.import_module(pipeline_module_name).get_pipeline(None)
    pipeline_info = pipeline.to_json()

    mq = RabbitMQ(
        connection = get_connection(**config["rabbitmq"]),
        topic = pipeline.id
    )
    # create rabbitmq topic, etc
    mq.initialize()

    with zkdb(**config['zookeeper']) as db:
        error = db.register_pipeline(pipeline.id, pipeline_namespace_name, pipeline_image_name, pipeline_module_name, pipeline_info)
        if error != ZKError.OK:
            print(f"register pipeline failed, error is: {error}")

def unregister_command(config:dict, pipeline_id:str):
    with zkdb(**config['zookeeper']) as db:
        error = db.unregister_pipeline(pipeline_id)
        if error != ZKError.OK:
            print(f"unregister pipeline failed, error is: {error}")

def list_command(config):
    with zkdb(**config['zookeeper']) as db:
        pipelines = db.get_pipelines()

    for pipeline in pipelines:
        print(f"{pipeline['info']['id']}:")
        print(f"    namespace: {pipeline['namespace_name']}")
        print(f"    image    : {pipeline['image_name']}")
        print(f"    module   : {pipeline['module']}")
        print(f"    running  : {'Yes' if pipeline['is_running'] else 'No'}")
        if len(pipeline["executors"]) == 0:
            print("    executors: None")
        else:
            print("    executors:")
            for executor in pipeline["executors"]:
                executor_info = executor["info"]
                print(f"        {executor_info['id']}:")
                print(f"            start_time            = {executor_info['start_time']}")
                print(f"            pid                   = {executor_info['pid']}")
                print(f"            generator_id          = {executor_info['generator_id']}")


def get_generator_ids(pipeline):
    generator_ids = []
    for node in pipeline["info"]["nodes"]:
        input_port_count = 0
        for port in node["ports"]:
            if port["type"] == "INPUT": # hack, better use PortType enum
                input_port_count += 1
                break
        if input_port_count == 0:
            generator_ids.append(node["id"])
    return generator_ids


def stop_command(config:dict, pipeline_id:str):
    with zkdb(**config['zookeeper']) as db:
        error, pipeline = db.get_pipeline(pipeline_id)

    if error != ZKError.OK:
        print(f"Unable to get pipeline, error is {error}")
        return

    if not pipeline['is_running']:
        print(f"Pipeline is not running!")
        return

    for generator_id in get_generator_ids(pipeline):
        name = f"firebird-pipeline--{pipeline_id}--{generator_id}"
        print(f"Delete statefulset: {name}")
        K8ACCESSOR.delete_statefulset(namespace=pipeline["namespace_name"], name=name)
        print()

    name = f"firebird-pipeline--{pipeline_id}"
    print(f"Delete statefulset: {name}")
    K8ACCESSOR.delete_deployment(namespace=pipeline["namespace_name"], name=name)

    with zkdb(**config['zookeeper']) as db:
        error = db.set_pipeline_is_running(pipeline_id, False)
        if error != ZKError.OK:
            print(f"Unable to change pipeline to stopped status, error is {error}")
    print()


def start_command(config, pipeline_id, replicas):
    with zkdb(**config['zookeeper']) as db:
        error, pipeline = db.get_pipeline(pipeline_id)
    
    if error != ZKError.OK:
        print(f"Unable to get pipeline, error is {error}")
        return

    if pipeline['is_running']:
        print(f"Pipeline is already running!")
        return

    rollback_is_running = False
    try:
        with zkdb(**config['zookeeper']) as db:
            error = db.set_pipeline_is_running(pipeline_id, True)
            if error != ZKError.OK:
                print(f"Unable to change pipeline to running status, error is {error}")
                return
        rollback_is_running = True
        K8ACCESSOR.apply(
            template_name="pipeline.yaml",
            context={
                "pipeline_namespace_name": pipeline["namespace_name"],
                "pipeline_image_name": pipeline["image_name"],
                "pipeline_id": pipeline_id,
                "replicas": replicas,
                "generator_ids": get_generator_ids(pipeline)
            }
        )
        rollback_is_running = False
    finally:
        if rollback_is_running:
            with zkdb(**config['zookeeper']) as db:
                error = db.set_pipeline_is_running(pipeline_id, False)
                if error != ZKError.OK:
                    print(f"Unable to rollback pipeline's is_running status, error is {error}")

